import os
from tornado.options import Error


def parse_environment(self, final=True):
    for name in os.environ:
        normalized = self._normalize_name(name)
        normalized = normalized.lower()
        if normalized in self._options:
            option = self._options[normalized]
            if option.multiple:
                if not isinstance(os.environ[name], (list, str)):
                    raise Error("Option %r is required to be a list of %s "
                                "or a comma-separated string" %
                                (option.name, option.type.__name__))

            if type(os.environ[name]) == str and option.type != str:
                option.parse(os.environ[name])
            else:
                option.set(os.environ[name])



from urllib import parse


def replace_query_param(url, key, val):
    (scheme, netloc, path, query, fragment) = parse.urlsplit(str(url))
    query_dict = parse.parse_qs(query, keep_blank_values=True)
    query_dict[str(key)] = [str(val)]
    query = parse.urlencode(sorted(list(query_dict.items())), doseq=True)
    return parse.urlunsplit((scheme, netloc, path, query, fragment))


def remove_query_param(url, key):
    (scheme, netloc, path, query, fragment) = parse.urlsplit(str(url))
    query_dict = parse.parse_qs(query, keep_blank_values=True)
    query_dict.pop(key, None)
    query = parse.urlencode(sorted(list(query_dict.items())), doseq=True)
    return parse.urlunsplit((scheme, netloc, path, query, fragment))



import sqlalchemy
from sqlalchemy import inspect, or_

from jet_bridge.filters.char_filter import CharFilter
from jet_bridge.filters.filter import EMPTY_VALUES
from jet_bridge.filters.filter_class import FilterClass


def filter_search_field(field):
    allowed_fields = [
        sqlalchemy.VARCHAR,
        sqlalchemy.TEXT,
        sqlalchemy.CHAR,
        sqlalchemy.JSON
    ]

    return isinstance(field.type, tuple(allowed_fields))


def get_model_search_filter(Model):
    mapper = inspect(Model)
    search_fields = list(map(lambda x: x, filter(filter_search_field, mapper.columns)))

    class SearchFilter(CharFilter):
        def filter(self, qs, value):
            value = self.clean_value(value)
            if value in EMPTY_VALUES:
                return qs

            operators = list(map(lambda x: x.ilike('%{}%'.format(value)), search_fields))
            return qs.filter(or_(*operators))

    return SearchFilter


def get_model_filter_class(Model):
    search_filter = get_model_search_filter(Model)

    class ModelFilterClass(FilterClass):
        _search = search_filter()

        class Meta:
            model = Model

    return ModelFilterClass



from collections import Mapping


class Field(object):
    field_name = None

    def __init__(self, *args, **kwargs):
        self.required = kwargs.pop('required', True)
        self.read_only = kwargs.pop('read_only', False)
        self.write_only = kwargs.pop('write_only', False)
        self.many = kwargs.pop('many', False)

    def validate(self, value):
        return value

    def get_value(self, data):
        try:
            if isinstance(data, Mapping):
                field_value = data[self.field_name]
            else:
                field_value = getattr(data, self.field_name)
        except (KeyError, AttributeError):
            return

        if not getattr(self, 'many', False) and isinstance(field_value, list):
            field_value = field_value[0]

        if isinstance(field_value, bytes):
            field_value = field_value.decode('utf8')

        return field_value

    def run_validation(self, value):
        value = self.to_internal_value(value)
        return value

    def to_internal_value_item(self, value):
        raise NotImplementedError

    def to_internal_value(self, value):
        if self.many:
            return list(map(lambda x: self.to_internal_value_item(x), value))
        else:
            return self.to_internal_value_item(value)

    def to_representation_item(self, value):
        raise NotImplementedError

    def to_representation(self, value):
        if self.many:
            return list(map(lambda x: self.to_representation_item(x), value))
        else:
            return self.to_representation_item(value)



import six

from jet_bridge.fields.field import Field


class CharField(Field):

    def __init__(self, *args, **kwargs):
        self.trim_whitespace = kwargs.pop('trim_whitespace', True)
        super().__init__(*args, **kwargs)

    def to_internal_value_item(self, value):
        if value is None:
            return
        value = six.text_type(value)
        return value.strip() if self.trim_whitespace else value

    def to_representation_item(self, value):
        if value is None:
            return
        return six.text_type(value)



from sqlalchemy import inspect
from sqlalchemy.orm.base import ONETOMANY

from jet_bridge.db import Session, MappedBase
from jet_bridge.models import data_types
from jet_bridge.permissions import HasProjectPermissions
from jet_bridge.responses.base import Response
from jet_bridge.serializers.model_description import ModelDescriptionSerializer
from jet_bridge.utils.db_types import map_data_type
from jet_bridge.views.base.api import APIView


class ModelDescriptionsHandler(APIView):
    serializer_class = ModelDescriptionSerializer
    permission_classes = (HasProjectPermissions,)
    session = Session()

    def get_queryset(self):
        non_editable = ['id']
        hidden = ['__jet__token']

        def map_column(column):
            params = {}
            data_type = map_data_type(str(column.type))

            if len(column.foreign_keys):
                foreign_key = next(iter(column.foreign_keys))
                data_type = data_types.FOREIGN_KEY
                params['related_model'] = {
                    'model': foreign_key.column.table.name
                }

            return {
                'name': column.name,
                'db_column': column.name,
                'field': data_type,
                'filterable': True,
                'editable': column.name not in non_editable,
                'params': params
            }

        def map_relation(relation):
            field = None
            through = None

            if relation.direction == ONETOMANY:
                field = 'ManyToOneRel'

            return {
                'name': relation.key,
                'related_model': {
                    'model': relation.table.name
                },
                'field': field,
                'related_model_field': relation.primaryjoin.right.name,
                'through': through
            }

        def map_table(cls):
            mapper = inspect(cls)
            name = mapper.selectable.name
            return {
                'model': name,
                'db_table': name,
                'fields': list(map(map_column, mapper.columns)),
                'hidden': name in hidden,
                'relations': list(map(map_relation, filter(lambda x: x.direction == ONETOMANY, mapper.relationships)))
            }

        return list(map(map_table, MappedBase.classes))

    def get(self, *args, **kwargs):
        queryset = self.get_queryset()
        serializer = self.serializer_class(instance=queryset, many=True)
        response = Response(serializer.representation_data)
        self.write_response(response)



from jet_bridge.db import Session
from jet_bridge.paginators.page_number import PageNumberPagination
from jet_bridge.views.base.api import APIView


class GenericAPIView(APIView):
    serializer_class = None
    filter_class = None
    pagination_class = PageNumberPagination
    _paginator = None
    lookup_field = 'id'
    lookup_url_kwarg = None
    session = Session()
    action = None

    def get_model(self):
        raise NotImplementedError

    def get_queryset(self):
        raise NotImplementedError

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field

        assert lookup_url_kwarg in self.path_kwargs

        model_field = getattr(self.get_model(), self.lookup_field)
        obj = queryset.filter(getattr(model_field, '__eq__')(self.path_kwargs[lookup_url_kwarg])).first()

        self.check_object_permissions(obj)

        return obj

    def get_filter(self, *args, **kwargs):
        filter_class = self.get_filter_class()
        if not filter_class:
            return
        kwargs['context'] = self.filter_context()
        return filter_class(*args, **kwargs)

    def get_filter_class(self):
        return self.filter_class

    def filter_context(self):
        return {
            'request': self.request,
            'handler': self
        }

    def filter_queryset(self, queryset):
        filter_instance = self.get_filter()
        if filter_instance:
            queryset = filter_instance.filter_queryset(queryset)
        return queryset

    @property
    def paginator(self):
        if not self._paginator:
            if self.pagination_class is None:
                self._paginator = None
            else:
                self._paginator = self.pagination_class()
        return self._paginator

    def paginate_queryset(self, queryset):
        if self.paginator is None:
            return None
        return self.paginator.paginate_queryset(queryset, self)

    def get_paginated_response(self, data):
        assert self.paginator is not None
        return self.paginator.get_paginated_response(data)

    def get_serializer(self, *args, **kwargs):
        serializer_class = self.get_serializer_class()
        kwargs['context'] = self.get_serializer_context()
        return serializer_class(*args, **kwargs)

    def get_serializer_class(self):
        return self.serializer_class

    def get_serializer_context(self):
        return {
            'request': self.request,
            'view': self,
            'session': self.session
        }



import requests

from jet_bridge import settings
from jet_bridge.db import Session
from jet_bridge.models.token import Token


def api_method_url(method):
    return '{}/{}'.format(settings.API_BASE_URL, method)


def register_token():
    session = Session()
    token = session.query(Token).first()

    if token:
        return token, False

    url = api_method_url('project_tokens/')
    headers = {
        'User-Agent': 'Jet Django'
    }

    r = requests.request('POST', url, headers=headers)
    success = 200 <= r.status_code < 300

    if not success:
        print('Register Token request error', r.status_code, r.reason)
        return None, False

    result = r.json()

    token = Token(token=result['token'].replace('-', ''), date_add=result['date_add'])
    session.add(token)
    session.commit()

    return token, True


def is_token_activated():
    session = Session()
    token = session.query(Token).first()

    if not token:
        return False

    url = api_method_url('project_tokens/{}/'.format(token.token))
    headers = {
        'User-Agent': 'Jet Django'
    }

    r = requests.request('GET', url, headers=headers)
    success = 200 <= r.status_code < 300

    if not success:
        return False

    result = r.json()

    return bool(result.get('activated'))


def reset_token():
    session = Session()
    session.query(Token).delete()
    session.commit()

    return register_token()


def project_auth(token, permission=None):
    session = Session()
    project_token = session.query(Token).first()

    if not project_token:
        return {
            'result': False
        }

    url = api_method_url('project_auth/')
    data = {
        'project_token': project_token.token,
        'token': token
    }
    headers = {
        'User-Agent': 'Jet Django'
    }

    if permission:
        data.update(permission)

    r = requests.request('POST', url, data=data, headers=headers)
    success = 200 <= r.status_code < 300

    if not success:
        print('Project Auth request error', r.status_code, r.reason)
        return {
            'result': False
        }

    result = r.json()

    if result.get('access_disabled'):
        return {
            'result': False,
            'warning': result.get('warning')
        }

    return {
        'result': True,
        'warning': result.get('warning')
    }



from jet_bridge.views.base.api import APIView


class NotFoundHandler(APIView):

    def write_error(self, status_code, **kwargs):
        self.set_status(404)
        self.finish('<h1>Not Found</h1><p>The requested URL {} was not found on this server.</p>'.format(self.request.uri))



from jet_bridge.filters.model import get_model_filter_class
from jet_bridge.permissions import HasProjectPermissions, ModifyNotInDemo
from jet_bridge.serializers.model import get_model_serializer
from jet_bridge.views.mixins.model import ModelAPIViewMixin
from jet_bridge.db import MappedBase


class ModelHandler(ModelAPIViewMixin):
    model = None
    permission_classes = (HasProjectPermissions, ModifyNotInDemo)

    @property
    def required_project_permission(self):
        return {
            'permission_type': 'model',
            'permission_object': self.path_kwargs['model'],
            'permission_actions': {
                'create': 'w',
                'update': 'w',
                'partial_update': 'w',
                'destroy': 'd',
                'retrieve': 'r',
                'list': 'r',
                'aggregate': 'r',
                'group': 'r',
                'reorder': 'w',
                'reset_order': 'w'
            }.get(self.action, 'w')
        }

    def get_model(self):
        if self.model:
            return self.model

        self.model = MappedBase.classes[self.path_kwargs['model']]

        return self.model

    def get_serializer_class(self):
        Model = self.get_model()
        return get_model_serializer(Model)

    def get_filter_class(self):
        return get_model_filter_class(self.get_model())

    def get_queryset(self):
        Model = self.get_model()

        return self.session.query(Model)



import tornado.web
from tornado.escape import json_decode


class APIView(tornado.web.RequestHandler):
    permission_classes = []

    @property
    def data(self):
        content_type = self.request.headers.get('Content-Type', '')
        if content_type.startswith('application/json'):
            return json_decode(self.request.body)
        else:
            return self.request.body_arguments

    def prepare(self):
        if self.request.method != 'OPTIONS':
            self.check_permissions()

    def set_default_headers(self):
        ACCESS_CONTROL_ALLOW_ORIGIN = 'Access-Control-Allow-Origin'
        ACCESS_CONTROL_EXPOSE_HEADERS = 'Access-Control-Expose-Headers'
        ACCESS_CONTROL_ALLOW_CREDENTIALS = 'Access-Control-Allow-Credentials'
        ACCESS_CONTROL_ALLOW_HEADERS = 'Access-Control-Allow-Headers'
        ACCESS_CONTROL_ALLOW_METHODS = 'Access-Control-Allow-Methods'

        self.set_header(ACCESS_CONTROL_ALLOW_ORIGIN, '*')
        self.set_header(ACCESS_CONTROL_ALLOW_METHODS, 'GET, POST, PUT, PATCH, DELETE, OPTIONS')
        self.set_header(ACCESS_CONTROL_ALLOW_HEADERS, 'Authorization,DNT,User-Agent,X-Requested-With,If-Modified-Since,Cache-Control,Content-Type,Range,X-Application-Warning')
        self.set_header(ACCESS_CONTROL_EXPOSE_HEADERS, 'Content-Length,Content-Range,X-Application-Warning')
        self.set_header(ACCESS_CONTROL_ALLOW_CREDENTIALS, 'true')

    def get_permissions(self):
        return [permission() for permission in self.permission_classes]

    def check_permissions(self):
        for permission in self.get_permissions():
            if not permission.has_permission(self):
                raise Exception(getattr(permission, 'message', None))

    def check_object_permissions(self, obj):
        for permission in self.get_permissions():
            if not permission.has_object_permission(self, obj):
                raise Exception(getattr(permission, 'message', None))

    def options(self, *args, **kwargs):
        self.set_status(204)
        self.finish()

    def write_response(self, response):
        for name, value in response.header_items():
            self.set_header(name, value)
        self.write(response.render())

    def write_error(self, status_code, **kwargs):
        print(kwargs.get('exc_info'))
        self.write('Error %s' % status_code)



from jet_bridge.models import data_types

map_data_types = [
    {'query': 'VARCHAR', 'operator': 'startswith', 'date_type': data_types.TEXT},
    {'query': 'TEXT', 'operator': 'equals', 'date_type': data_types.TEXT},
    {'query': 'BOOLEAN', 'operator': 'equals', 'date_type': data_types.BOOLEAN},
    {'query': 'INTEGER', 'operator': 'equals', 'date_type': data_types.INTEGER},
    {'query': 'SMALLINT', 'operator': 'equals', 'date_type': data_types.INTEGER},
    {'query': 'NUMERIC', 'operator': 'startswith', 'date_type': data_types.FLOAT},
    {'query': 'VARCHAR', 'operator': 'startswith', 'date_type': data_types.DATE_TIME},
    {'query': 'TIMESTAMP', 'operator': 'startswith', 'date_type': data_types.TIMESTAMP},
]
default_data_type = data_types.TEXT


def map_data_type(value):
    for rule in map_data_types:
        if rule['operator'] == 'equals' and value == rule['query']:
            return rule['date_type']
        elif rule['operator'] == 'startswith' and value[:len(rule['query'])] == rule['query']:
            return rule['date_type']
    return default_data_type



from jet_bridge import settings
from jet_bridge.utils.backend import project_auth


class BasePermission(object):

    def has_permission(self, view):
        return True

    def has_object_permission(self, view, obj):
        return True


class HasProjectPermissions(BasePermission):
    token_prefix = 'Token '
    project_token_prefix = 'ProjectToken '

    def has_permission(self, view):
        token = view.request.headers.get('Authorization')
        permission = getattr(view, 'required_project_permission', None)

        if not token:
            return False

        if token[:len(self.token_prefix)] == self.token_prefix:
            token = token[len(self.token_prefix):]

            result = project_auth(token, permission)

            if result.get('warning'):
                view.headers['X-Application-Warning'] = result['warning']

            return result['result']
        elif token[:len(self.project_token_prefix)] == self.project_token_prefix:
            token = token[len(self.project_token_prefix):]

            result = project_auth(token, permission)

            if result.get('warning'):
                view.headers['X-Application-Warning'] = result['warning']

            return result['result']
        else:
            return False


class ModifyNotInDemo(BasePermission):

    def has_permission(self, view):
        if not settings.READ_ONLY:
            return True
        if view.action in ['create', 'update', 'partial_update', 'destroy']:
            return False
        return True



from jet_bridge.fields import field, CharField
from jet_bridge.filters import lookups

EMPTY_VALUES = ([], (), {}, '', None)


class Filter(object):
    field_class = field
    lookup_operators = {
        lookups.EXACT: {'operator': '__eq__'},
        lookups.ICONTAINS: {'operator': 'ilike', 'post_process': lambda x: x.format('%{}%')},
        lookups.IN: {'operator': 'in_', 'field_class': CharField, 'field_kwargs': {'many': True}, 'pre_process': lambda x: x.split(',')},
    }

    def __init__(self, field_name=None, model=None, lookup=lookups.DEFAULT_LOOKUP):
        self.field_name = field_name
        self.name = field_name
        self.model = model
        self.lookup = lookup

    def clean_value(self, value):
        if value is None:
            return
        return self.field_class().to_internal_value(value)

    def apply_lookup(self, qs, lookup, value):
        model_field = getattr(self.model, self.field_name)
        lookup_operator = self.lookup_operators[lookup]
        operator = lookup_operator['operator']
        pre_process = lookup_operator.get('pre_process')
        post_process = lookup_operator.get('post_process')
        field_class = lookup_operator.get('field_class')
        field_kwargs = lookup_operator.get('field_kwargs')

        if pre_process:
            value = pre_process(value)

        if field_class:
            value = field_class(**field_kwargs).to_internal_value(value)

        if post_process:
            value = post_process(value)

        return qs.filter(getattr(model_field, operator)(value))

    def filter(self, qs, value):
        value = self.clean_value(value)
        if value in EMPTY_VALUES:
            return qs
        return self.apply_lookup(qs, self.lookup, value)




class Router(object):
    routes = [
        {
            'path': '',
            'method_mapping': {
                'get': 'list',
                'post': 'create'
            },
        },
        {
            'path': '(?P<id>[^/]+)/',
            'method_mapping': {
                'get': 'retrieve',
                'put': 'update',
                'patch': 'partial_update',
                'delete': 'destroy'
            }
        }
    ]
    urls = []

    def register(self, prefix, viewset):
        for route in self.routes:
            actions = route['method_mapping']
            actions = dict(filter(lambda x: hasattr(viewset, x[1]), actions.items()))

            if len(actions) == 0:
                continue

            class ActionHandler(viewset):
                pass

            for method, method_action in actions.items():
                def create_action_method(action):
                    def action_method(inner_self, *args, **kwargs):
                        inner_self.action = action
                        return getattr(inner_self, action)(*args, **kwargs)
                    return action_method

                func = create_action_method(method_action)
                setattr(ActionHandler, method, func)

            url = '{}{}'.format(prefix, route['path'])
            self.urls.append((url, ActionHandler))



from jet_bridge.permissions import HasProjectPermissions
from jet_bridge.responses.base import Response
from jet_bridge.serializers.sql import SqlSerializer
from jet_bridge.views.base.api import APIView


class SqlHandler(APIView):
    permission_classes = (HasProjectPermissions,)

    def post(self):
        serializer = SqlSerializer(data=self.data)
        serializer.is_valid(raise_exception=True)

        response = Response(serializer.execute())
        self.write_response(response)



from jet_bridge import status
from jet_bridge.responses.base import Response


class DestroyAPIViewMixin(object):

    def delete(self, *args, **kwargs):
        instance = self.get_object()
        self.perform_destroy(instance)
        self.write_response(Response(status=status.HTTP_204_NO_CONTENT))

    def perform_destroy(self, instance):
        self.session.delete(instance)
        self.session.commit()



from collections import OrderedDict

import math
import tornado.web

from jet_bridge.paginators.pagination import Pagination
from jet_bridge.responses.base import Response
from jet_bridge.utils.http import replace_query_param, remove_query_param


class PageNumberPagination(Pagination):
    page_number = None
    page_size = 20
    page_query_param = 'page'
    page_size_query_param = '_per_page'
    max_page_size = 10000
    handler = None

    def paginate_queryset(self, queryset, handler):
        page_number = self.get_page_number(handler)
        if not page_number:
            return None

        page_size = self.get_page_size(handler)
        if not page_size:
            return None

        self.count = queryset.count()
        self.page_number = page_number
        self.page_size = page_size
        self.handler = handler

        queryset = queryset.offset((page_number - 1) * page_size).limit(page_size)

        return list(queryset)

    def get_pages_count(self):
        return int(math.ceil(self.count / self.page_size))

    def get_paginated_response(self, data):
        return Response(OrderedDict([
            ('count', self.count),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data),
            ('num_pages', self.get_pages_count()),
            ('per_page', self.page_size),
        ]))

    def get_page_number(self, handler):
        try:
            result = int(handler.get_argument(self.page_query_param))
            return max(result, 1)
        except (tornado.web.MissingArgumentError, ValueError):
            return 1

    def get_page_size(self, handler):
        if self.page_size_query_param:
            try:
                result = int(handler.get_argument(self.page_size_query_param))
                result = max(result, 1)

                if self.max_page_size:
                    result = min(result, self.max_page_size)

                return result
            except (tornado.web.MissingArgumentError, ValueError):
                pass

        return self.page_size

    def has_next(self):
        return self.page_number < self.get_pages_count()

    def has_previous(self):
        return self.page_number > 1

    def next_page_number(self):
        return self.page_number + 1

    def previous_page_number(self):
        return self.page_number - 1

    def get_next_link(self):
        if not self.has_next():
            return None
        url = self.handler.request.full_url()
        page_number = self.next_page_number()
        return replace_query_param(url, self.page_query_param, page_number)

    def get_previous_link(self):
        if not self.has_previous():
            return None
        url = self.handler.request.full_url()
        page_number = self.previous_page_number()
        if page_number == 1:
            return remove_query_param(url, self.page_query_param)
        return replace_query_param(url, self.page_query_param, page_number)



from jet_bridge import VERSION
from jet_bridge.responses.base import Response
from jet_bridge.views.base.api import APIView


class ApiHandler(APIView):

    def get(self):
        response = Response({
            'version': VERSION,
            'type': 'jet_bridge'
        })
        self.write_response(response)



from datetime import datetime

import tornado.ioloop
import tornado.web
import jet_bridge.adapters.postgres

from jet_bridge import settings, VERSION
from jet_bridge.router import Router
from jet_bridge.utils.backend import is_token_activated
from jet_bridge.views.api import ApiHandler
from jet_bridge.views.main import MainHandler
from jet_bridge.views.message import MessageHandler
from jet_bridge.views.model import ModelHandler
from jet_bridge.views.model_description import ModelDescriptionsHandler
from jet_bridge.views.not_found import NotFoundHandler
from jet_bridge.views.register import RegisterHandler
from jet_bridge.views.sql import SqlHandler


def make_app():
    router = Router()

    router.register('/api/models/(?P<model>[^/]+)/', ModelHandler)

    return tornado.web.Application([
        (r'/', MainHandler),
        (r'/register/', RegisterHandler),
        (r'/api/', ApiHandler),
        (r'/api/register/', RegisterHandler),
        (r'/api/model_descriptions/', ModelDescriptionsHandler),
        (r'/api/sql/', SqlHandler),
        (r'/api/messages/', MessageHandler),
    ] + router.urls, debug=settings.DEBUG, default_handler_class=NotFoundHandler)


def main():
    app = make_app()
    app.listen(settings.PORT, settings.ADDRESS)
    address = 'localhost' if settings.ADDRESS == '0.0.0.0' else settings.ADDRESS
    url = 'http://{}:{}/'.format(address, settings.PORT)

    print(datetime.now().strftime('%B %d, %Y - %H:%M:%S %Z'))
    print('Jet Bridge version {}'.format(VERSION))
    print('Starting server at {}'.format(url))

    if settings.DEBUG:
        print('Server is running in DEBUG mode')

    print('Quit the server with CONTROL-C')

    if not is_token_activated():
        print('[!] Your server token is not activated')
        print('[!] Go to {}register/ to activate'.format(url))

    tornado.ioloop.IOLoop.current().start()

if __name__ == '__main__':
    main()
